import { AuthState } from "./enum/AuthState";
import { Endian, createReader, createWriter } from "bufferstuff";
import { existsSync, readFileSync } from "fs";
import { Socket } from "net";
import { Packet } from "./enum/Packet";
import { WebSocketServer } from "ws";
import IServerConfig from "./interface/IServerConfig";
import Constants from "./Constants";

export default class Server {
	constructor() {
		if (!existsSync("./config/server-config.json")) {
			console.error("server-config.json is missing!");
			process.exit(1);
		}
		const config:IServerConfig = JSON.parse(readFileSync("./config/server-config.json").toString());
		const server = new WebSocketServer({ port: config.port }, () => console.log(`Server started at ${config.port}`));
	
		server.on("connection", (socket) => {
			console.log("Connection");
			let queuedMessages = new Array<Buffer>();
			let connectedToServer = false;
			let connectingToServer = false;
			let authed = false;
			let client:Socket | null = null;
			const clientKeepAlive = setInterval(() => {
				socket.send(Constants.KEEPALIVE_PACKET);
			}, 5000);
	
			socket.on("message", (data, isBinary) => {
				if (!isBinary) {
					return;
				}
	
				// NOTE: The types declarations for ws are really messed up >:(
				// @ts-ignore
				const packetData = createReader(Endian.LE, data);
				const packetId = packetData.readUByte();
				let tempReadLength = 0;
				
				switch (packetId) {
					case Packet.KeepAlive: break; // We don't really care, it's just so CF doesn't drop the connection.
					case Packet.Auth:
						try {
							if (packetData.readShortString() !== config.authKey) {
								socket.send(createWriter(Endian.LE, 2).writeUByte(Packet.Auth).writeUByte(AuthState.Bad).toBuffer());
								socket.close();
								return;
							}
							socket.send(createWriter(Endian.LE, 2).writeUByte(Packet.Auth).writeUByte(AuthState.Good).toBuffer());
							authed = true;
		
							client = new Socket();
		
							client.on("connect", () => {
								connectedToServer = true;
								if (queuedMessages.length > 0) {
									for (const message of queuedMessages) {
										console.log("Sent", message);
										client?.write(message);
									}
									queuedMessages.length = 0;
								}
							});
		
							client.on("data", (chunk) => {
								socket.send(createWriter(Endian.LE, 5)
									.writeUByte(Packet.Data)
									.writeUInt(chunk.length)
									.writeBuffer(chunk)
									.toBuffer());
							});
	
							function clientCloseOrError() {
								socket.close();
							}
							client.on("close", clientCloseOrError);
							client.on("error", clientCloseOrError);
		
							connectingToServer = true;
							client.connect({
								host: config.localHost,
								port: config.localPort
							});
						} catch (e) {
							client?.end();
							client = null;
							socket.close();
						}
						break;
					case Packet.Data:
						if (!authed) {
							return;
						}
	
						tempReadLength = packetData.readUInt();
						if (connectedToServer) {
							client?.write(packetData.readBuffer(tempReadLength));
						} else {
							queuedMessages.push(packetData.readBuffer(tempReadLength));
						}
						//console.log("[SERVER] Data:", data, " Length:", tempReadLength);
						break;
				}
			});
	
			function closeOrError() {
				clearInterval(clientKeepAlive);
				client?.end();
			}
			socket.on("close", closeOrError);
			socket.on("error", closeOrError);
		});	
	}
}